// $Id: ParseHelper.java,v 1.18 2003/03/12 21:31:10 drager Exp $

package org.daml.validator;

import java.io.*;
import java.util.*;
import java.net.*;

//import org.xml.sax.SAXException;
//import org.w3c.rdf.model.*;
//import org.w3c.rdf.syntax.*;
//import org.w3c.rdf.util.*;

import com.hp.hpl.mesa.rdf.jena.model.Resource;
import com.hp.hpl.mesa.rdf.jena.model.Property;
import com.hp.hpl.mesa.rdf.jena.model.RDFNode;
import com.hp.hpl.mesa.rdf.jena.model.Statement;
import com.hp.hpl.mesa.rdf.jena.model.Model;
import com.hp.hpl.mesa.rdf.jena.model.Literal;
import com.hp.hpl.mesa.rdf.jena.model.StmtIterator;
import com.hp.hpl.mesa.rdf.jena.model.RDFException;
import com.hp.hpl.mesa.rdf.jena.model.NodeIterator;
import com.hp.hpl.mesa.rdf.jena.mem.ModelMem;

import org.daml.validator.indications.*;
import org.daml.validator.model.*;
import org.daml.validator.*;
import org.daml.validator.DAMLDatatypeValidator;
import org.daml.validator.xsd.DatatypeValidator;
import org.daml.validator.xsd.UnknownSchemaException;
import org.daml.validator.xsd.SchemaValidationException;


public class ParseHelper {

    public boolean VERBOSE = false;
    
    /**
     * URIs to be loaded
     */
    private Stack todo = new Stack();

    /**
     * URIs that have been loaded
     */
    public Vector loaded = new Vector();

    /**
     * URIs that had errors during loading
     */
    public Vector errors = new Vector();

    /**
     * URIs that were actually loaded
     */
    public Vector actuallyLoaded = new Vector();

    /**
     * Reference Stack for URIs in recurse mode
     */
    public Stack recurseStack = new Stack();

    /**
     * Flag denoting if referenced namespaces should be loaded.
     */
    public boolean recurseNS = false;

    /**
     * The number of levels deep to recursively load
     * -1 means infinite.
     */
    public int maxRecurseNS = 1;

    /**
     * Flag denoting if imported ontologies should be loaded.
     */
    public boolean recurseImports = true;
    
    /**
     * The number of levels deep to recursively load imported files.
     * -1 means infinite.
     */
    public int maxRecurseImports = -1;

    /**
     * Indications generated during parsing.
     */
    public Vector indications = new Vector();

    /**
     * Preference Object
     */
    Preferences pref;

    /**
     * Flag for having the namespace parser to print namespace indications as they are encountered.
     */
    public boolean printwarnings = false;

    /**
     * List of Predicates with the DAML imports semantics.
     */
    public Vector importsPredicates = new Vector();

    /**
     * XML Schema Datatype Validator 
     */
    DAMLDatatypeValidator datatypeValidator;

    public ParseHelper(Preferences preferences, Vector importsPredicates, DAMLDatatypeValidator datatypeValidator)
    {
	pref = preferences;
	recurseNS = pref.getBooleanProperty("recurseNS", recurseNS);
	recurseImports = pref.getBooleanProperty("recurseImports", recurseImports);
	maxRecurseNS = pref.getIntProperty("maxRecurseNS", maxRecurseNS);
	maxRecurseImports = pref.getIntProperty("maxRecurseImports", maxRecurseImports);
	this.importsPredicates = importsPredicates;
	this.datatypeValidator = datatypeValidator;
    }


    /**
     * Parses a URI into a model.  If recurse is true, it recursively fetches
     * all referenced namespace URIs.  I wrote this because the SiRPAC fetchSchemas
     * flag didn't work as I expected it to.
     */
    public void parseURI(String initialURI, Model model, boolean recurseNS)
	throws RDFException
    {
	boolean rNS = this.recurseNS;
 	this.recurseNS = recurseNS;
 	todo.push(initialURI);
 	helper(model);
	this.recurseNS = rNS;
    }

    /**
     * Parses a list of URIs into a model.
     */
    public void parseURIs(Collection URIs, Model model, boolean recurseNS)
	throws RDFException
    {
	boolean rNS = this.recurseNS;
	this.recurseNS = recurseNS;
	// Only add URIs that end with #
	addNamespaces(URIs);
	helper(model);
	this.recurseNS = rNS;
    }
   
    /**
     * Converts the namespace reference stack to a string for indication messages.
     */
    private static String stackString(Stack stack)
    {
	String loc = "";
	for (int i = stack.size(); i > 0; ) {
	    loc += stack.elementAt(--i);
	    if (i > 0) { loc += " <- ";}
	}
	return(loc);
    }


    
    /**
     * Implements the recursive parser using a depth first algorithm (implemented
     * iteratively.  If an error occurs it is able to produce the stack of URIs
     * that lead the the problem namespace.
     */
    private void helper(Model model) 
	throws RDFException
    {
	// Used to keep track of nodes that referenced the current namespace
	Stack stack = recurseStack;

	HashMap map = new HashMap();

	while (!todo.empty()) {
	    String uri = (String)todo.pop();
	    if (uri == null) {
		stack.pop();
	    } else {
		// Each URI gets its own model so I only have to search that model for more URIs
		Model m = new ValidatorModelMem(false);
		
		String loadedURI = loadURI(uri, m, stack, map);

		todo.push(null);
		stack.push(loadedURI);

		if (VERBOSE) System.out.println("\nSTACK:"+stack.size()+"\n"+stack+"\n");

		// Copy all the statements to the model
		for (StmtIterator en = m.listStatements(); en.hasNext(); ) {
		    Statement s = (Statement)en.next();
		    model.add(s);
		}

		// Determine what files to import
		if (recurseImports &&
		    (maxRecurseImports < 0 || stack.size() < maxRecurseImports+1)) {
		    if (VERBOSE) System.out.println("CHECKING IMPORTS "+stack.size());
		    Set set = getImportedURIs(m);
		    if (VERBOSE) System.out.println(set);
		    addNamespaces(set);
		}

		// Determine what referenced namespaces to load
		if (recurseNS &&
		    (maxRecurseNS < 0 || stack.size() < maxRecurseNS+1) ) {
		    Set set = getReferencedURIs(m);
		    //HashSet set = new HashSet();
		    //RDFUtil u = new RDFUtil();
		    //u.collectNamespaces(m, set);
		    addNamespaces(set);
		}
	    }
	}
	if (!map.isEmpty()) {
	    indications.add(new CacheIndication(map));
	}
    }


    /**
     * Load a uri, resolving it with the cache/outdatedURIs if necessary.
     */
    private String loadURI(String uri, Model m, Stack stack, Map map) 
	throws RDFException
    {
	// Test the uri for a file type extension and issue a warning if exists
	// TURNING OFF FOR NOW
	// testNamespace(uri, stack);

	// This bit of code takes the uri, normalizes it (so we can
	// compare it to other URIs), then determines if there is a
	// forced substitution and/or cache file to load.  Either way,
	// we need the parser to think that it is loading the original
	// (non-normalized) uri, otherwise the namespace won't match
	// and we'll get lots of extra undefined resource
	// inidications.

	String loadURI = uri;
	String normalizedURI = pref.normalizeURI(uri);
	// This checks for outdated URIs (according to preference file) and 
	// generates indications and possibly changes what to load.
	String resolveURI = resolveURI(normalizedURI, stack);
	String cacheURI = cacheURI(resolveURI);
	if (!cacheURI.equals(normalizedURI)) {
	    // If there was a forced substitution or cache file to load
	    // then don't load the original.
	    loadURI = cacheURI;
	}
		
	try {
	    //System.out.println("\n>>>>>\n"+uri+"\n"+loadURI+"\n"+cacheURI);
	    // Load if it hasn't already been loaded
	    if (!loaded.contains(loadURI) &&
		!errors.contains(loadURI) &&
		!datatypeValidator.hasNamespace(loadURI) ) {
		
		if (loadURI.equals(uri)) {
		    System.out.println("\nLoading URI "+uri);
		} else {
		    System.out.println("\nLoading URI "+loadURI+"\ninstead of "+uri);
		    map.put(uri, loadURI);
		}

		// Check to see if URI is an XML Schema file
		boolean isSchema = datatypeValidator.isASchema(loadURI);
		//System.out.println("isASchema("+loadURI+") = "+isSchema);
		if (isSchema) {
		    // Add the original uri, not the cache file.  Will check for cache again when 
		    // we load the file.
		    // Exceptions are caught below
		    datatypeValidator.addFromCache(uri, loadURI);

		} else {  // A (presumably) RDF/DAML file

		    // If a cache file is being loaded, trick the parser into thinking
		    // it is the original file.
		    java.net.URL url = new URL (normalizeURI(loadURI));
		    java.io.Reader reader = new BufferedReader(new InputStreamReader(url.openStream()));
		    m.read(reader, uri);

// 		    m.setSourceURI(uri);
// 		    org.xml.sax.InputSource source = new org.xml.sax.InputSource( url.openStream() );
// 		    source.setSystemId(uri);
// 		    parser.parse(source, new ModelConsumer(m));
		
		    loaded.add(loadURI);
		    loaded.add(uri);
		    actuallyLoaded.add(loadURI);
		}
	    }
	} catch (RDFException se) {
	    errors.add(loadURI);
	    Indication i = new ParseIndication(loadURI, stack, se);
	    indications.add(i);
	    if (printwarnings) {
		System.out.println();
		System.out.println("RDFException: "+i);
		System.out.println("Will attempt to proceed with validation...\n");
	    }
	}  catch (FileNotFoundException fe) {
	    errors.add(loadURI);
	    String loc = uri+" referenced by "+stackString(stack);
	    Indication i = new InvalidNamespaceIndication(loadURI, stack, fe);
	    indications.add(i);
	    if (printwarnings) {
		System.out.println();
		System.out.println("FileNotFoundException: "+i);
		System.out.println("Will attempt to proceed with validation...\n");
	    }
	} catch (java.net.MalformedURLException me) {
	    errors.add(loadURI);
	    Indication i = new InvalidNamespaceIndication(loadURI, stack, me);
	    indications.add(i);
	    if (printwarnings) {
		System.out.println();
		System.out.println("MalformedURLException: "+i);
		System.out.println("Will attempt to proceed with validation...\n");
	    }
	} catch (UnknownSchemaException use) {
	    errors.add(loadURI);
	    Indication i = new InvalidNamespaceIndication(loadURI, stack, use);
	    indications.add(i);
	    if (printwarnings) {
		System.out.println();
		System.out.println("UnknownSchemaException: "+i);
		System.out.println("Will attempt to proceed with validation...\n");
	    }
	} catch (SchemaValidationException sve) {
	    errors.add(loadURI);
	    Indication i = new ParseIndication(loadURI, stack, sve);
	    indications.add(i);
	    if (printwarnings) {
		System.out.println();
		System.out.println("SchemaValidationException: "+i);
		System.out.println("Will attempt to proceed with validation...\n");
	    }
	} catch (java.io.IOException ie) {
	    errors.add(loadURI);
	    Indication i = new ParseIndication(loadURI, stack, ie);
	    indications.add(i);
	    if (printwarnings) {
		System.out.println();
		System.out.println("IOException: "+i);
		System.out.println("Will attempt to proceed with validation...\n");
	    }
	} catch (com.hp.hpl.mesa.rdf.jena.model.RDFError rdfe) {
	    errors.add(loadURI);
	    Indication i = new ParseIndication(loadURI, stack, rdfe);
	    indications.add(i);
	    if (printwarnings) {
		System.out.println();
		System.out.println("RDFError: "+i);
		System.out.println("Will attempt to proceed with validation...\n");
	    }
	}
	return(loadURI);
    }

    /**
     * Checks a URI to see if it is outdated or stored in the cache.
     * Returns the URI to load.
     */
    private String resolveURI(String uri, Stack stack) 
    {
	String result = uri;
	if (pref.outdatedURIs.hasSubstitution(result)) {
	    String sub = pref.outdatedURIs.getSubstitution(result);
	    // URI is outdated
	    if (pref.outdatedURIs.getWarn(result)) {
		// Issue a warning for this URI
		indications.add(new OutdatedURIIndication(result, stack, sub));
	    }
	    if (pref.outdatedURIs.getLoad(result)) {
		result = sub;
	    }
	}
	// Return the actual URI to load
	return(result);
    }

    private String cacheURI(String uri)
    {
	String result = uri;
	// Check the cache file
	if (pref.cache.containsKey(uri)) {
	    // There exists a cache file for this uri.
	    result = (String)pref.cache.get(result);
	}
	return(result);
    }
	

    private void addNamespaces(Collection set) 
    {
	for (Iterator i = set.iterator(); i.hasNext(); ) {
	    String namespace = (String)i.next();
	    if (!todo.contains(namespace) &&
	       !loaded.contains(namespace) &&
	       !errors.contains(namespace)) {
		if (namespace.endsWith("#")) {
		    if (VERBOSE) System.out.println("DEBUG: Adding URI "+namespace);
		    todo.push(namespace);
		} else {
		    if (VERBOSE) System.out.println("DEBUG: Found namespace not ending in #.  "+namespace);
		}
	    }	
	}
    }

    /**
     * Returns a set of URIs referenced by the model.  
     * Does not include explicitly imported URIs.
     */
    public Set getReferencedURIs(Model m)
	throws RDFException
    {
	HashSet set = new HashSet();
	String ns;

	// get namespaces
	for (StmtIterator e = m.listStatements(); e.hasNext(); ) {
	    Statement s = (Statement)e.next();
	    ns = getNamespace(s.getSubject());
	    if (ns != null) set.add(ns);
	    ns = getNamespace(s.getPredicate());
	    if (ns != null) set.add(ns);
	    RDFNode node = s.getObject();
	    if (node instanceof Resource) {
		ns = getNamespace((Resource)node);
		if(ns != null) set.add(ns);
	    }
	}
	return(set);
    }


    /**
     * Returns a set of URIs imported by the model.
     * Bug: This does *not* check the subject of the imports statement
     * to make sure it is the current document.
     */
    public Set getImportedURIs(Model m)
	throws RDFException
    {
	HashSet set = new HashSet();
	String ns;

	// get imported files
	for (StmtIterator e = m.listStatements(); e.hasNext(); ) {
	    Statement s = (Statement)e.next();
  	    if (importsPredicates.contains(s.getPredicate())) {
		RDFNode ob = s.getObject();
		String uri = ob.toString();
		// Add a # to the end if necessary so we will load it.
		if (!uri.endsWith("#")) {
		    set.add(uri+"#");
		} else {
		    set.add(uri);
		}
	    }
	}
	return(set);
    }


    public static String getNamespace (Resource r) 
    {
	if (r == null) return null;
	String uri = r.getURI();
	if (uri == null) return null;
	int posn = uri.indexOf('#');
	if (posn >= 0) {
	    return uri.substring(0, posn+1);
	} else {
	    return uri+"#";
	}
    }

    /**
     * Tests URI for the existance of a filename extension.
     */
    public void testNamespace(String uri, Stack stack) 
    {	
	int dotIndex = uri.lastIndexOf('.');
	int slashIndex = uri.lastIndexOf('/');
	int hashIndex = uri.lastIndexOf('#');

	if ((dotIndex == -1) || (dotIndex < slashIndex)) {
	} else {
	    String ext=uri.substring(dotIndex);
	    if (ext.endsWith("#")) {
		ext = ext.substring(0, ext.length()-1);
	    }
	    indications.add(new FileExtensionIndication(uri, stack, ext));
	}
    }

    private String normalizeURI(String uri) {

	// normalise uri
	URL url = null;
	try {
	    url = new URL (uri);
	} catch (Exception e) {
	    try {
		if(uri.indexOf(':') == -1)
		    url = new URL ("file", null, uri);
	    } catch (Exception e2) {}
	}
	return url != null ? url.toString() : uri;
    }


}
