//$Id: URISubstitution.java,v 1.2 2001/08/15 18:51:50 drager Exp $

package org.daml.validator;

import java.util.HashMap;
import java.util.Iterator;

/**
 * Store and process URI substitutions.
 * Each original URI is associated with a replacement URI (or null) and
 * a boolean used to signal a warning.
 */

public class URISubstitution 
{

    private HashMap uriHash;
    private HashMap warnHash;
    private HashMap loadHash;


    /**
     * Generate a new URI Substitution set.
     */
    public URISubstitution()
    {
	uriHash = new HashMap();
	warnHash = new HashMap();
	loadHash = new HashMap();
    }

    public String toString()
    {
	return("[URISubstitution: "+uriHash+" "+warnHash+" "+loadHash+"]");
    }

    /**
     * Add a substitution to the set.
     */
    public void put(String originalURI, String newURI, boolean warn, boolean load) 
    {
	uriHash.put(originalURI, newURI);
	warnHash.put(originalURI, new Boolean(warn));
	loadHash.put(originalURI, new Boolean(load));
    }

    /**
     * Determine if a URI has a substitution.
     */
    public boolean hasSubstitution(String originalURI)
    {
	return(uriHash.containsKey(originalURI));
    }

    /**
     * Get the substitution URI for the original URI.
     */
    public String getSubstitution(String originalURI) 
    {
	return((String)uriHash.get(originalURI));
    }

    /**
     * Determine if a warning needs to be generated for this URI.
     */
    public boolean getWarn(String originalURI)
    {
	Boolean tmp = (Boolean)warnHash.get(originalURI);
	if (tmp == null) {
	    return false;
	} else {
	    return(tmp.booleanValue());
	}
    }

    /**
     * Determine if the replacement URI should be loaded instead.
     */
    public boolean getLoad(String originalURI)
    {
	Boolean tmp = (Boolean)loadHash.get(originalURI);
	if (tmp == null) {
	    return false;
	} else {
	    return(tmp.booleanValue());	
	}
    }

    /**
     * Returns an Iterator over the URIs that have substitutions.
     */
    public Iterator iterator() 
    {
	return(uriHash.keySet().iterator());
    }
    
}
