// $Id: Validator.java,v 1.50 2003/03/12 21:31:10 drager Exp $

package org.daml.validator;

import java.io.*;
import java.util.*;

// import org.xml.sax.SAXException;
// import org.w3c.rdf.model.*;
// import org.w3c.rdf.syntax.*;
// import org.w3c.rdf.util.*;
// import org.w3c.rdf.implementation.syntax.sirpac.*;

import com.hp.hpl.mesa.rdf.jena.model.*;
import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.mesa.rdf.jena.mem.ModelMem;

import org.daml.validator.indications.*;
import org.daml.validator.model.*;
import org.daml.validator.*;
import org.daml.validator.DAMLDatatypeValidator;
import org.daml.validator.xsd.DatatypeValidator;
import org.daml.validator.xsd.UnknownSchemaException;
import org.daml.validator.xsd.SchemaValidationException;
import org.daml.validator.xsd.UnknownSimpleTypeException;
import org.w3c.dom.*;
import org.apache.xerces.dom.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.*;

// Used by test function only
import org.daml.validator.xsd.imp.*;
import oracle.xml.parser.schema.*;

public class Validator {

    boolean DEBUG = false;

    /**
     * Preference File
     */
    String preferenceFile = "preferences.xml";

    /**
     * Preference Object
     */
    Preferences preferences;

    /**
     * Property arguments
     */
    Vector properties = new Vector();

    /**
     * URI to validate.
     */
    String validateURI;

    /**
     * Is set to true if arguments are parsed correctly.
     */
    private boolean start = false;

    /**
     * List of indications that are found.
     */
    Vector indications = new Vector();

    /**
     * The URI of the current file being validated.
     */
     String currentURI;

    /**
     * The model for the URI being validated
     */
    Model validateModel;
    
    /**
     * Model of all statements to be validated, and all statements from referenced namespaces.
     */
    DAMLModelImpl overallModel;

    /**
     * Text output desired.
     */
    private boolean textOutput = true;

    /**
     * XML output desired, and stream to use if so.
     */
    private java.io.PrintWriter xmlOutput = null;
    private File xmlOutputFile = null;

    /**
     * Text file output desired, and stream to use if so.
     */
    private java.io.PrintWriter txtOutput = null;

    /**
     * Version of the validator
     */
    public static String version = "20030312";

    /**
     * The maximum number of indications to show.
     */
    public int maxIndications = 100;

    /**
     * List of all URIs that have been loaded
     */
    Vector loadedURIs;

    /**
     * List of all Namespaces that have been loaded
     */
    Vector loadedNamespaces;

    /**
     * List of all URIs that encounted errors while parsing
     */
    Vector errorURIs;

    /**
     * XML Schema Datatype Validator
     */
    DAMLDatatypeValidator datatypeValidator = new DAMLDatatypeValidator();

    /**
     * URLs to always load
     */
    private Set defaultImports;



    public static void usage()
    {
	System.out.println("Usage:  <arguments> <uri-or-file>");
	System.out.println(" arguments:");
        System.out.println(" -[no]text         show results as text (default is -text, unless -xml, or -txt is specified)" );
	System.out.println(" -txt <txtFile>    store results in <txtFile>" );
	System.out.println(" -xml <xmlFile>    store results in <xmlFile>" );
	System.out.println(" -level <indication-class> {error|warning|information}");
	System.out.println("                   set the default level for this class of indication.");
	System.out.println(" -report <int>     limit the number of reported indications."); 
	System.out.println(" -pref <filename>  specify preferences file.");
	System.out.println(" -<prop>:<val>     set property <prop> to <val>, overriding the preference file.");
    }

    /**
     * Main method for running the Validator as an application.
     */
    public static void main (String args[]) 
	throws Exception
    {
	// Set up system to use jsse for https processing
	//System.setProperty("java.protocol.handler.pkgs", "com.sun.net.ssl.internal.www.protocol"); 
	//java.security.Security.addProvider(new com.sun.net.ssl.internal.ssl.Provider());	

	Validator validator = new Validator(args);
	if (validator.start) {
	    //validator.test();
	    validator.run();
	} else {
	    validator.usage();
	}
    }
    
    /**
     * Parses the arguments and collects the URIs to validate
     */
    private void parseArguments(String [] args) {
	int i = 0;
	start = true;
	// parse arguments
	
	try {
	    while (args[i].startsWith("-") && start) {
		if (args[i].equals("-text")) {
		    textOutput = true;
		    i++;
		} else if (args[i].equals("-notext")) {
		    textOutput = false;
		    i++;
		} else if (args[i].equals("-xml")) {
		    try {
			xmlOutput = new java.io.PrintWriter(new java.io.FileOutputStream(args[1+i]));
			xmlOutputFile = new File(args[i+1]);
			textOutput = false;
			i += 2;
		    } catch (ArrayIndexOutOfBoundsException ex) {
			start = false;
		    } catch (java.io.FileNotFoundException ex) {
			System.err.println(ex);
			start = false;
		    }
		} else if (args[i].equals("-txt")) {
		    try {
			txtOutput = new java.io.PrintWriter(new java.io.FileOutputStream(args[1+i]));
			textOutput = false;
			i += 2;
		    } catch (ArrayIndexOutOfBoundsException ex) {
			start = false;
		    } catch (java.io.FileNotFoundException ex) {
			System.err.println(ex);
			start = false;
		    }
		} else if (args[i].equals("-level")) {
		    try {
			start = changeIndicationLevel(args[1+i], args[2+i]);
			i += 3;
		    } catch (ArrayIndexOutOfBoundsException ex) {
			start = false;
		    }
		} else if (args[i].equals("-report")) {
		    try {
			System.out.println(i);
			System.out.println(args);
			maxIndications = Integer.parseInt(args[i+1]);
			i += 2;
		    } catch (NumberFormatException ex) {
			start = false;
		    }
		} else if (args[i].equals("-pref")) {
		    preferenceFile = args[i+1];
		    i += 2;
		} else if (args[i].indexOf(":") > 0) {
		    properties.add(args[i]);
		    System.out.println("Property Argument: "+args[i]);
		    i += 1;
		} else {
		    start = false;
		}
	    }
	} catch (ArrayIndexOutOfBoundsException ex) {
	    start = false;
	}

	// Add URIs to validate
	// Only doing one URI at a time now
	if (start && (i < args.length)) {
	    validateURI = args[i++];
	} else {
	    start = false;
	}
    }

    /**
     * Changes the default level for a class of Indications.  
     * Returns true if indication and level are valid.
     */
    private boolean changeIndicationLevel(String indication, String level) 
    {
	Class ind;
	try {
	    ind = Class.forName(indication);
	} catch (ClassNotFoundException ex) {
	    System.out.println("Class "+indication+" not found.");
	    return(false);
	}

	//System.out.println(ind.getFields());

	int l = Indication.ERROR;
	if (level.equals("error")) {
	    l = Indication.ERROR;
	} else if (level.equals("warning")) {
	    l = Indication.WARNING;
	} else if (level.equals("information")) {
	    l = Indication.INFORMATION;
	} else {
	    System.out.println("Indication level " + level + " not valid.");
	    return(false);
	}


	try {
	    ind.getField("defaultLevel").setInt(null,l);
	    //System.out.println("Set Default level for "+indication+" to "+level);
	} catch (NoSuchFieldException fe) {
	    System.out.println("Class "+indication+" does not have a public default level.");
	    return(false);
	} catch (IllegalAccessException ee) {
	    System.out.println("Can not change the default level for indication "+indication);
	    return(false);
	}

	return(true);
    }


    /**
     * Create Validator and parse arguments.
     */
    public Validator(String[] args) {
	super();
	parseArguments(args);
	if (start) {
	    // Load Preferences
	    try {
		preferences = Preferences.parse(preferenceFile);
	    } catch (Exception e) {
		System.out.println("An error occurred while reading preference file "+preferenceFile);
		System.out.println(e.toString());
		System.out.println();
		start = false;
	    }
	    // set command line properties in the preferences object
	    setArgProperties(preferences, properties);
	}
    }

    public Validator(String preferenceFile) 
	throws Exception 
    {
	super();
	this.preferenceFile = preferenceFile;
	preferences = Preferences.parse(preferenceFile);
    }

    public Validator() 
	throws Exception
    {
	super();
	preferences = Preferences.parse(preferenceFile);
    }

    /**
     * Sets property values from the command line in the preference object.
     * properties is a Collection of "-name:value" strings.
     */
    private void setArgProperties(Preferences preferences, Collection properties)
    {
	Iterator i = properties.iterator();
	while ( i.hasNext() ) {
	    // parse out name and value of the property
	    Object obj = i.next();
	    if (obj instanceof String) {
		String prop = (String)obj;
		int pos = prop.indexOf(":");
		if (prop.charAt(0) == '-' && pos > 1 && (pos+1) < prop.length()) {
		    String name = prop.substring(1, pos);
		    String value = prop.substring(pos+1);
		    preferences.setProperty(name, value);
		}
	    }
	}
    }


    /**
     * Initialize the validator by clearing out any previously loaded models.
     */
    public void initialize()
    {
	// List of all URIs that have been loaded
	loadedURIs = new Vector();

	// List of all Namespaces that have been loaded
	loadedNamespaces = new Vector();

	// List of all URIs that encounted errors while parsing
	errorURIs = new Vector();

	// Reset the XML Schema Datatype Validator
	datatypeValidator.initialize();
	
	// Reset the Indication counter
	Indication.resetIdCounter();

	// Reset the UndefinedResourceIndication cache
	UndefinedResourceIndication.reset();

	// List of indications
	indications = new Vector();
	
	// Clear models
	validateModel = null;

	// Initialize and Set up equivalence/subclass/domain/range semantics for the DAMLModel.
	overallModel = new DAMLModelImpl(datatypeValidator, preferences);
	defaultImports = new HashSet();
	defaultImports.add("http://www.daml.org/2001/03/daml+oil#");
	defaultImports.add("http://www.w3.org/1999/02/22-rdf-syntax-ns#");
	defaultImports.add("http://www.w3.org/2000/01/rdf-schema#");
	defaultImports.add("http://www.w3.org/2000/10/XMLSchema#");
	initializeDAMLModel(overallModel);

	// Reset the URI being parsed
 	currentURI = null;
    }

    /**
     * Resets the indications and clears the backing model used to validate
     * against.
     */
    public void reset()
    {
	// Reset the overall Model
	// Initialize and Set up equivalence/subclass/domain/range semantics for the DAMLModel.
	overallModel = new DAMLModelImpl(datatypeValidator, preferences);
	initializeDAMLModel(overallModel);

	// Reset Indication list
	indications.clear();
	
	// Reset the Indication counter
	Indication.resetIdCounter();

	// Reset the UndefinedResourceIndication cache
	UndefinedResourceIndication.reset();

    }
    
    /**
     * Starts the validation process.
     */
    public void run() 
	throws Exception 
    {
	
	//================================================================
	// Initialize
	//================================================================
	initialize();

	//================================================================
	// Generate the Model that we will validate
	//================================================================

	// For now, only do one URI
	String uri = toNamespace(validateURI);
	System.out.println("\nloading "+uri+" to validate.");

	validateModel = new ValidatorModelMem(true);

// 	validateModel = factory.createModel();
// 	parser = (SiRPAC)factory.createParser();
// 	parser.fetchSchemas(false);

	// Parse the URI to validate.  This model is used as a list of statements to validate.
	boolean validateURIerror = false;
	try {
	    validateModel.read(uri);
// 	    RDFUtil.parse(uri, parser, validateModel);
	    loadedURIs.add(uri);
	    indications.addAll(((ValidatorModelMem)validateModel).getIndications());
	} catch (RDFException se) {
	    errorURIs.add(uri);
	    Indication i = new ParseIndication(uri, null, se);
	    indications.add(i);
	    validateURIerror=true;
	} catch (RDFError re) {
	    errorURIs.add(uri);
	    Indication i = new ParseIndication(uri, null, re);
	    indications.add(i);
	    validateURIerror=true;
// 	} catch (FileNotFoundException fe) {
// 	    errorURIs.add(uri);
// 	    Indication i = new InvalidNamespaceIndication(uri, null, fe);
// 	    indications.add(i);
// 	    validateURIerror=true;
// 	} catch (java.io.IOException ie) {
// 	    errorURIs.add(uri);
// 	    Indication i = new ParseIndication(uri, null, ie);
// 	    indications.add(i);
// 	    validateURIerror=true;
 	}
	//================================================================
	// Determine if there are any statements in the validate model
	// If not, then there is no need to create the backing model.
	//================================================================
	if (validateModel.size() == 0) {
	    validateURIerror=true;
	    Indication i = new NotRDFIndication(uri);
	    indications.add(i);
	}

	if (!validateURIerror) {

	    //================================================================
	    // Generate the Model to validate against
	    //================================================================

	    // Parse URI again, this time recursively loading referenced URIs.
	    System.out.println("Reading referenced namespaces...");
	    Model m = new ModelMem();
// 	    Model m = factory.createModel();
// 	    parser = (SiRPAC)factory.createParser();	
	    // NOTE: this seems broken!  will need to load referenced namespaces myself
// 	    parser.fetchSchemas(true);

	    ParseHelper p = new ParseHelper(preferences, overallModel.importsPredicates, datatypeValidator);

	    // Create a Set of all namespaces/imports that are referenced
	    // Use the recurseNS and recurseImports values
	    Set set = new HashSet();
	    if (p.recurseNS && (p.maxRecurseNS < 0 || p.maxRecurseNS > 0))
		set.addAll(p.getReferencedURIs(validateModel));
	    if (p.recurseImports && (p.maxRecurseImports < 0 || p.maxRecurseImports > 0))
		set.addAll(p.getImportedURIs(validateModel));

	    // Add default imports namespaces
	    set.addAll(defaultImports);

	    // Make sure we don't load the URI we are validating (yet)
	    p.loaded.add(uri);
	    p.recurseStack.push(uri);
	    set.remove(uri);

	    // Don't load these known offenders:
	    p.loaded.add("http://www.w3.org/TR/WD-rdf-schema#");
	    p.loaded.add("http://www.w3.org/XML/xmlschema#");
	    p.loaded.add("http://www.w3c.org/xml/xmlschema#");

	    //System.out.println("\nLoading "+set);
	    p.parseURIs(set, m, true);
	    loadedURIs = p.loaded;
	    indications.addAll(p.indications);
	    indications.add(new LoadIndication(p.actuallyLoaded, p.recurseNS, p.maxRecurseNS, p.recurseImports, p.maxRecurseImports));

	    // Determine which version of the DAML spec is being used,
	    // and set the DAMLVocabulary used to generate new
	    // statements.
	    if (loadedURIs.contains(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.baseURI)) {
		overallModel.setDAMLVocabulary("org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL");
	    } else if (loadedURIs.contains(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.baseURI)) {
		overallModel.setDAMLVocabulary("org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL");
	    } else if (loadedURIs.contains(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.baseURI)) {
		overallModel.setDAMLVocabulary("org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL");
	    }

		

	    // Tell user when there were parse errors during load.
	    if (!p.errors.isEmpty()) {
		System.out.println("\nErrors occurred while attempting to load the following namespaces.  Will continue to validate, but results may not be correct.");
		for(Enumeration e = p.errors.elements(); e.hasMoreElements(); ) {
		    Object obj = e.nextElement();
		    System.out.println(obj);
		    errorURIs.add(obj);
		}
		System.out.println();
	    }

	    // Now place the RDF model into the DAML Model
	    overallModel.addModel(m);

	    //================================================================
	    // Generate datatype statements in the Model for all datatypes
	    //================================================================
	    Set simpleTypes = datatypeValidator.getSimpleTypeSet();
	    //System.err.println("Adding Datatype statements: "+simpleTypes);
	    overallModel.addDatatypes(simpleTypes);


	    //================================================================
	    // Validate the model against the overall model
	    //================================================================

	    validate(validateModel, uri);

	}

	// Generate the Results
	collectIndications();
	outputIndications();
    }


    /**
     * Gathers certain types of indications together.
     */
    private void collectIndications() 
    {
    }


    
    private void outputIndications () 
	throws RDFException
    {
	if (xmlOutput != null) {
	    outputXMLIndications ();
	}
	if (txtOutput != null) {
	    outputTxtIndications();
	    txtOutput.close();
	}
	if (textOutput) {
	    outputTextIndications ();
	}
    }

    /**
     * Note: This ignores the maxIndications restriction.
     */
    private void outputTxtIndications() {
	txtOutput.println(validateURI);
	txtOutput.println();
	// Sort the Indications
	TreeSet sorted = new TreeSet();
	for (Iterator i = indications.iterator(); i.hasNext(); ) {
	    Indication indication = (Indication)i.next();
	    sorted.add(indication.toComparableString());
	}

	
	// Print the Indications
	for (Iterator i = sorted.iterator(); i.hasNext(); ) {
	    txtOutput.println((String)i.next());
	    txtOutput.println("\n");
	}	
    }
	
    private void outputTextIndications ()
    {
	System.out.println("\n=====================================");
	System.out.println("The following Indications were found:\n");

	// Print the Indications
	int count = 0;
	for (Iterator i = indications.iterator(); (i.hasNext() && (count < maxIndications));) {
	    count++;
	    Indication indication = (Indication)i.next();
	    System.out.println(indication);
	    System.out.println("\n");
	}
	
	int size = indications.size();
	if (count < size) {
	    System.out.println("Plus " + (size-count) + " Indications not shown.");
	}	    
    }

    private void outputXMLIndications ()
    throws RDFException
    {
	xmlOutput.println("<?xml version=\"1.0\"?>");
	xmlOutput.println("<daml-validator>");
	xmlOutput.println("<version>" + version + "</version>");
	xmlOutput.println("<document>" + validateURI + "</document>");
	xmlOutput.println("<results>");
	xmlOutput.println("<date>"+(new Date())+"</date>");
	xmlOutput.println("<total>"+indications.size()+"</total>");

	int count = indications.size();
	if (count > maxIndications)
	    count = maxIndications;
	xmlOutput.println("<displayed>"+count+"</displayed>");

	// Print the Indications
	count = 0;
	for (Iterator i = indications.iterator(); (i.hasNext() && (count < maxIndications));) {
	    count++;
	    Indication indication = (Indication)i.next();
	    xmlOutput.println(indication.toXML());
	}	
	xmlOutput.println("</results>");
	xmlOutput.println("</daml-validator>");
	xmlOutput.close();
    }

    /**
     * Associate predicates with the semantics of the DAML model
     */
    public static void initializeDAMLModel (DAMLModelImpl m)
    {
	//defaultImports.add("file:cache/additions#");

	//m.datatypeRestrictionClasses.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.DatatypeRestriction);
	//m.datatypeRestrictionClasses.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.DatatypeRestriction);
	//m.equivalentPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.sameIndividualAs);
	//m.objectRestrictionClasses.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.ObjectRestriction);
	//m.objectRestrictionClasses.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.ObjectRestriction);
	//m.propertyClasses.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.DatatypeProperty);
	//m.propertyClasses.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.DatatypeProperty);
	//m.propertyClasses.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.ObjectProperty);
	//m.propertyClasses.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.ObjectProperty);

	m.cardinalityPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.cardinality);
	m.cardinalityPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.cardinality);
	m.cardinalityPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.cardinality);

	m.cardinalityQPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.cardinalityQ);
	m.cardinalityQPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.cardinalityQ);
	m.cardinalityQPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.cardinalityQ);

	m.classClasses.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.Class);
	m.classClasses.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.Class);
	m.classClasses.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.Class);
	m.classClasses.add(com.hp.hpl.mesa.rdf.jena.vocabulary.RDFS.Class);

	m.complementOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.complementOf);
	m.complementOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.complementOf);
	m.complementOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.complementOf);

	m.datatypePropertyClasses.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.DatatypeProperty);
	m.datatypePropertyClasses.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.DatatypeProperty);

	m.datatypeClasses.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.Datatype);
	m.datatypeClasses.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.Datatype);

	m.domainPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.domain);
	m.domainPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.domain);
	m.domainPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.domain);
	m.domainPredicates.add(com.hp.hpl.mesa.rdf.jena.vocabulary.RDFS.domain);

	m.equivalentPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.equivalentTo);
	m.equivalentPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.sameClassAs);
	m.equivalentPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.samePropertyAs);
	m.equivalentPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.equivalentTo);
	m.equivalentPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.equivalentTo);
	m.equivalentPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.sameClassAs);
	m.equivalentPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.sameClassAs);
	m.equivalentPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.sameIndividualAs);
	m.equivalentPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.sameIndividualAs);
	m.equivalentPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.samePropertyAs);
	m.equivalentPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.samePropertyAs);

	m.equivalentToPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.equivalentTo);
	m.equivalentToPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.equivalentTo);
	m.equivalentToPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.equivalentTo);

	m.hasClassPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.hasClass);
	m.hasClassPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.hasClass);
	m.hasClassPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.hasClass);

	m.hasClassQPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.hasClassQ);
	m.hasClassQPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.hasClassQ);
	m.hasClassQPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.hasClassQ);

	m.hasValuePredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.hasValue);
	m.hasValuePredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.hasValue);
	m.hasValuePredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.hasValue);

	m.importsPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.imports);
	m.importsPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.imports);
	m.importsPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.imports);

	m.intersectionOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.intersectionOf);
	m.intersectionOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.intersectionOf);
	m.intersectionOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.intersectionOf);

	m.literalClasses.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.Literal);
	m.literalClasses.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.Literal);
	m.literalClasses.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.Literal);
	m.literalClasses.add(com.hp.hpl.mesa.rdf.jena.vocabulary.RDFS.Literal);

	m.maxCardinalityPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.maxCardinality);
	m.maxCardinalityPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.maxCardinality);
	m.maxCardinalityPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.maxCardinality);

	m.maxCardinalityQPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.maxCardinalityQ);
	m.maxCardinalityQPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.maxCardinalityQ);
	m.maxCardinalityQPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.maxCardinalityQ);

	m.minCardinalityPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.minCardinality);
	m.minCardinalityPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.minCardinality);
	m.minCardinalityPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.minCardinality);

	m.minCardinalityQPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.minCardinalityQ);
	m.minCardinalityQPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.minCardinalityQ);
	m.minCardinalityQPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.minCardinalityQ);

	m.objectPropertyClasses.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.ObjectProperty);
	m.objectPropertyClasses.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.ObjectProperty);

	m.onPropertyPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.onProperty);
	m.onPropertyPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.onProperty);
	m.onPropertyPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.onProperty);

	m.oneOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.oneOf);
	m.oneOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.oneOf);
	m.oneOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.oneOf);

	m.propertyClasses.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.Property);
	m.propertyClasses.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.Property);
	m.propertyClasses.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.Property);
	m.propertyClasses.add(com.hp.hpl.mesa.rdf.jena.vocabulary.RDF.Property);

	m.rangePredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.range);
	m.rangePredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.range);
	m.rangePredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.range);
	m.rangePredicates.add(com.hp.hpl.mesa.rdf.jena.vocabulary.RDFS.range);

	m.resourceClasses.add(com.hp.hpl.mesa.rdf.jena.vocabulary.RDFS.Resource);

	m.restrictionClasses.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.Restriction);
	m.restrictionClasses.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.Restriction);
	m.restrictionClasses.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.Restriction);

	m.samePropertyAsPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.samePropertyAs);
	m.samePropertyAsPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.samePropertyAs);
	m.samePropertyAsPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.samePropertyAs);

	m.subClassOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.subClassOf);
	m.subClassOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.subClassOf);
	m.subClassOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.subClassOf);
	m.subClassOfPredicates.add(com.hp.hpl.mesa.rdf.jena.vocabulary.RDFS.subClassOf);

	m.subPropertyOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.subPropertyOf);
	m.subPropertyOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.subPropertyOf);
	m.subPropertyOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.subPropertyOf);
	m.subPropertyOfPredicates.add(com.hp.hpl.mesa.rdf.jena.vocabulary.RDFS.subPropertyOf);

	m.ancestorPredicates.addAll(m.subClassOfPredicates);
	m.ancestorPredicates.addAll(m.subPropertyOfPredicates);

	m.toClassPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.toClass);
	m.toClassPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.toClass);
	m.toClassPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.toClass);

	m.typePredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.type);
	m.typePredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.type);
	m.typePredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.type);
	m.typePredicates.add(com.hp.hpl.mesa.rdf.jena.vocabulary.RDF.type);

	m.unionOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.unionOf);
	m.unionOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.unionOf);
	m.unionOfPredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.unionOf);

	m.valuePredicates.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.value);
	m.valuePredicates.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.value);
	m.valuePredicates.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.value);
	m.valuePredicates.add(com.hp.hpl.mesa.rdf.jena.vocabulary.RDF.value);

	m.listClasses.add(org.daml.rdf.vocabulary.daml_oil_20010111.DAML_OIL.List);
	m.listClasses.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.List);
	m.listClasses.add(org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL.List);

	m.allPropertyClasses.addAll(m.propertyClasses);
	m.allPropertyClasses.addAll(m.objectPropertyClasses);
	m.allPropertyClasses.addAll(m.datatypePropertyClasses);
    }


    // Convert a filename to a URI and ensure URIs end in #
    private String toNamespace(String str) 
    {
	if (str.startsWith("http:") && (str.lastIndexOf('#') != -1)) {
	    return (str);
	} else if (str.startsWith("file:") && (str.lastIndexOf('#') != -1)) {
	    return (str);
	} else {
	    if (str.lastIndexOf('#') == -1) {
		str += "#";
	    }
	    if (!str.startsWith("file:") && !str.startsWith("http:")) {
		str = "file:"+str;
	    }
	    return(str);
	}
    }

    private String resolvePath(String path)
    {
	try {
	File file = new File(path);
	System.err.println(">>>>>>>>");
	System.err.println(file.getAbsolutePath());
	System.err.println(file.getCanonicalPath());
	System.err.println(">>>>>>>>");
	} catch (IOException e) {
	}
	return path;
    }

    /*
    private void test(Model m)
    throws RDFException 
    {

	System.out.println("DEBUG: IN TEST FUNCTION");

	// Loop through the statements
	for (Enumeration e = m.elements(); e.hasMoreElements();) {
	    edu.stanford.db.rdf.model.i.StatementImpl s = (edu.stanford.db.rdf.model.i.StatementImpl)e.nextElement();
	    System.out.println("-"+s);
	    System.out.println(" "+s.getSourceURI());
	}
    }
    */

    /*
    private void test()
	throws Exception
    {
	System.out.println("");
	System.out.println("================================================================");
	System.out.println(" BEGINNING TEST FUNCTION ");
	System.out.println("================================================================");
	System.out.println("");

	// What do the datatype definition objects look like?
	SimpleTypeRegistry str = ((DatatypeValidatorImp)(datatypeValidator.datatypeValidator)).simpleTypeRegistry;
	XSDSimpleType st = str.getSimpleType("file:/daml/tools/validator/cache/www.daml.org_2001_03_daml+oil-ex-dt", "clothingsize");
	System.out.println("XSDSimpleType is "+st);
	System.out.println("getBase() = "+st.getBase());
	System.out.println("getBasicType() = "+st.getBasicType());
	String v = st.getVariety();
	System.out.println("getVariety() = "+v);
	if (v.equals("union")) {
	    //System.out.println("members = "+st.members);
	} else if (v.equals("item")) {
	    //System.out.println("item = "+st.item);
	} 	

	System.out.println("");
	System.out.println("================================================================");
	System.out.println(" END OF TEST FUNCTION ");
	System.out.println("================================================================");
	System.out.println("");
    }
    */

    /*
    private void test()
    throws RDFException 
    {

	System.out.println("DEBUG: IN TEST FUNCTION");

	// Loop through the statements
	for (Enumeration e = validateModel.elements(); e.hasMoreElements();) {
	    Statement s = (Statement)e.nextElement();
	    Resource subject = s.subject();
	    Resource predicate = s.predicate();
	    RDFNode object = s.object();
	    DAMLModelImpl.Node sNode = overallModel.getNode(subject);
	    DAMLModelImpl.Node oNode = null;
	    if (object instanceof Resource) {
		overallModel.getNode((Resource)object);
	    }
	    DAMLModelImpl.Node pNode = overallModel.getNode(predicate);
	    
	    
	    // Do something
	    // Test the getRestriction functions
	    if (sNode != null) {
		//sNode.getRestrictions(false);
		System.out.println();
		Set set = sNode.getRestrictions(false);
		for (Iterator i = set.iterator(); i.hasNext(); System.out.println(((DAMLModelImpl.Node)i.next()).getRestrictions(false)) ) {}
	    }
	}
	    
    }
    */

    /**
     * Validates the statements in model checking for domain and range violations.
     */
    private void validateStatements(Model model, DAMLModelImpl overallModel)
	throws RDFException 
    {
	boolean DEBUG = false;
	System.out.println("Validating Model Statements...");

// 	currentURI = model.getSourceURI();

	// Loop through all statements in the data model and test:
	// 1. Domain types
	// 2. Range types
	for (StmtIterator e = model.listStatements(); e.hasNext() && (Indication.indicationCount() < maxIndications);) {
	    Statement s = (Statement)e.next();
	    Resource subject = s.getSubject();
	    Resource predicate = s.getPredicate();
	    RDFNode object = s.getObject();

	    if (DEBUG) System.out.println();
	    if (DEBUG) System.out.println(s);
	    
	    // lookup ontology Node for the predicate
	    DAMLModelImpl.Node predNode = overallModel.getNode(predicate);

	    if (predNode == null) {
		// Create an indication for this
		// NOTE: This is properly tested at the resource validation level.
		//Indication ind = UndefinedResourceIndication.create(predicate, s);
		//if (ind != null) { indications.add(ind); }
		if (DEBUG) System.out.println("No node for predicate found in model.");

	    } else {

		// Test the predicate to see if it is Property
		if (!predNode.isInstanceOfResources(overallModel.allPropertyClasses)) {
		    System.err.println("ERR: Didn't expect to get here -- Property "+predicate+" is not a property");
		    indications.add(new StatementIndication(s, "The property of this statement has not been defined as a Property."));
		}

		// get the domain of the predicate
		Vector domain = predNode.getDomain();

		// get the range of the predicate
		Vector range = predNode.getRange();

		// Test Domain
		if (domain.isEmpty()) {
		    // no domain restriction for the predicate
		    if (DEBUG) System.out.println("No domain restrictions.");
		    if (DEBUG) System.out.println("Domain statements: "+predNode.getDomainStatements());
		} else {
		    // validate the statement domain...
		    if (DEBUG) System.out.println("Domain restrictions: "+domain);

		    // get the subject
		    DAMLModelImpl.Node sNode = overallModel.getNode(subject);
		    //DAMLModelImpl.Node sType = sNode.getType();

		    if (subject instanceof Statement) {
			// Reitification - do nothing (yet)
			if (DEBUG) {System.out.println("DEBUG: Subject is Statement: "+subject);}
		    } else {
			// Node must be of all the domain types
			for (Enumeration en = domain.elements(); en.hasMoreElements(); ) {
			    DAMLModelImpl.Node d = (DAMLModelImpl.Node)en.nextElement();
			    // Compare class types
			    if (sNode.isInstanceOf(d)) {
				//System.out.println("found matching domain: "+sType+" "+domain);
			    } else if (overallModel.listClasses.contains(d.resource)) {
				// We can't handle List types yet, so ignore this.
				//System.out.println("!! FOUND LIST RESOURCE !!");
			    } else {
				indications.add(new IncorrectDomainIndication(s, d, sNode.getTypes()));
			    }
			}
		    }
		} // Done testing domain

		// New Code to test Range, separating datatype and objects
		if (range.isEmpty()) {
		    // No range restrictions for predicate
		    if (DEBUG) System.out.println("No Range restriction.");
		    if (DEBUG) System.out.println("Range statements "+predNode.getRangeStatements());
		} else {
		    // For each range restriction
		    if (DEBUG) System.out.println("Range restrictions: "+range);
		    for (Enumeration rangeEnum = range.elements(); rangeEnum.hasMoreElements(); ) {
			DAMLModelImpl.Node r = (DAMLModelImpl.Node)rangeEnum.nextElement();
			// Statement object must be an instance of the restriction
			if (overallModel.isInstanceOf(object, r)) {
			    // Good
			} else if (overallModel.listClasses.contains(r.resource)) {
			    // We can't handle List types yet, so ignore this.
			    //System.out.println("!! FOUND LIST RESOURCE !!");
			} else {
			    // Bad
			    if (object instanceof DAMLModelImpl.Node) {
				Set types = ((DAMLModelImpl.Node)object).getTypes();
				if (types.isEmpty()) {
				    indications.add(new IncorrectRangeIndication(s, r, null));
				} else {
				    indications.add(new IncorrectRangeIndication(s, r, types));
				}
			    } else {
				if (DEBUG) System.out.println("object class is "+object.getClass());
				Set tmpTypes2 = null;
				if (object instanceof Resource) {
				    DAMLModelImpl.Node tmpNode = overallModel.getNode((Resource)object);
				    Set tmpTypes = tmpNode.getTypes();
				    if (!tmpTypes.isEmpty()) {
					tmpTypes2 = tmpTypes;
				    }
				}
				indications.add(new IncorrectRangeIndication(s, r, tmpTypes2));
			    }
			}
		    }
		}

	    }
	}

    }
    
    /** 
     * Validate all referenced resources in the model to make sure they exist.
     */
    private void validateReferencedResources(Model validateModel, DAMLModelImpl overallModel)
	throws RDFException
    {
	System.out.println("Validating Referenced External Resources...");
	boolean DEBUG = false;

	String localURI = currentURI;
	if (DEBUG) System.out.println("localURI = "+localURI);

	for(StmtIterator en = validateModel.listStatements(); en.hasNext() && (Indication.indicationCount() < maxIndications);) {
	    Statement s = (Statement)en.next();
	    //System.out.println("DEBUG: statement "+s);
	    Resource predicate = s.getPredicate();
	    //DAMLModelImpl.Node node = overallModel.getNode(predicate);
	    //System.out.println();
	    //System.out.println(s);
	    //	    if (node != null) {node.getCardinality();}
	    if (DEBUG) System.out.println(predicate+" = "+isValidExternalResource(predicate,overallModel, localURI));
	    if (!isValidExternalResource(predicate, overallModel, localURI)) {
		// Assume that this is an undefined property and add it to the model
		if (DEBUG) System.out.println("Making undefined property explicit in model: "+predicate);
		overallModel.defineGenericProperty(predicate);
		indications.add(new UndefinedPropertyIndication(predicate, s));
	    }	
	    Resource subject = s.getSubject();
	    if (DEBUG) System.out.println(subject +" = "+isValidExternalResource(subject, overallModel, localURI));
	    if (!isValidExternalResource(subject, overallModel, localURI)) {
		Indication ind = UndefinedResourceIndication.create(subject, s);
		if (ind != null) {
		    indications.add(ind);
		}
	    }
	    RDFNode n = s.getObject();
	    if (n instanceof Resource) {
		if (DEBUG) System.out.println((Resource)n+" = "+isValidExternalResource((Resource)n,overallModel, localURI));
		if (!isValidExternalResource((Resource)n, overallModel, localURI)) {
		    // This is to get around problems with the name of imported URIs
		    if (overallModel.importsPredicates.contains(predicate)) {
			// This is an imported URI, so it probably won't match, so skip
		    } else if ((overallModel.onPropertyPredicates.contains(predicate)) ||
			       (overallModel.subPropertyOfPredicates.contains(predicate)) ||
			       (overallModel.samePropertyAsPredicates.contains(predicate)) )  {
			// This is an undefined property?
			if (DEBUG) System.out.println("Making undefined property explicit in model: "+predicate);
			overallModel.defineGenericProperty(predicate);
		    } else {
			Indication ind = UndefinedResourceIndication.create((Resource)n, s);
			if (ind != null) {
			    indications.add(ind);
			}
		    }
		}
	    }
	}
    }

    /**
     * Find invalid external resources.
     * Given a resource, it returns false if the resource is external the the localURI and
     * is not in the model m.  Statements, resources without namespaces, and local resources
     * always return true.
     */
    private boolean isValidExternalResource(Resource r, DAMLModelImpl m, String localURI) 
	throws RDFException
    {
	boolean DEBUG = false;
	if (DEBUG) {
	    System.out.println("\nDEBUG: Looking for "+r);
// 	    System.out.println("Label     "+r.getLabel());
	    System.out.println("Namespace "+r.getNameSpace());
	    System.out.println("LocalName "+r.getLocalName());
	    System.out.println("URI       "+r.getURI());
	}

	DAMLModelImpl.Node n = overallModel.getNode(r);
	if (n == null) {
	    if (DEBUG) System.out.println("DEBUG: No node for "+r);
	    // No such resource, but check to see if it is in the local namespace...
	    if (r instanceof Statement) {
		// Reitification
		return(true);
	    } else {
		String ns = ParseHelper.getNamespace(r);
		if (ns == null) {
		    // Doesn't have a namespace 
		    if (DEBUG) {System.out.println("DEBUG: Resource has no Namespace: "+r);}
		    return(true);
		} else if (ns.startsWith(localURI)) {
		    // if it is a local resource, then the node will definitely exist
		    // in the current model.
		    if (DEBUG) { System.out.println("DEBUG: local resource"); }
		    return(true);
		} else if (datatypeValidator.hasNamespace(ns)) {
		    // This is now dead code because all simple types are added as classes in the overall model.
		    return datatypeValidator.isAKnownSimpleType(ns, r.getLocalName());
		} else {
		    if (DEBUG) {
			System.out.println("DEBUG: No such resource "+r);
			System.out.println(ns);
			System.out.println(localURI);
		    }
		    return(false);
		}
	    }
	} else {
	    if (n.getTypeStatements().isEmpty()) {
		if (isGenid(r, localURI)) {
		    // Local genid### Resources without types are generated by RDF-API from <comment/> 
		    // This suppresses generating an error.
		    return(true);
		} else if ((r.getNameSpace() != null) && (r.getNameSpace().startsWith(localURI))) {
		    // Sometimes local URIs are refered to by external files.
		    // TBD - which of the above also apply here? reification?
		    return(true);
		} else {
		    return false;
		}
	    } else {
		return true;
	    }
	}
    }

    private void validateInternalResources(Model validateModel, DAMLModelImpl overallModel)
	throws RDFException
    {
	boolean DEBUG = false;
	System.out.println("Validating Referenced Internal Resources...");
	
	String localURI = currentURI;

	for(StmtIterator en = validateModel.listStatements(); en.hasNext() && (Indication.indicationCount() < maxIndications);) {
	    Statement s = (Statement)en.next();
	    Resource predicate = s.getPredicate();
	    if (!isValidInternalResource(predicate, overallModel, localURI)) {
		// Assume that this is an undefined property and add it to the model
		if (DEBUG) System.out.println("Making undefined property explicit in model: "+predicate);
		overallModel.defineGenericProperty(predicate);
		indications.add(new UndefinedPropertyIndication(predicate, s));
	    }	
	    Resource subject = s.getSubject();
	    if (!isValidInternalResource(subject, overallModel, localURI)) {
		Indication ind = UndefinedResourceIndication.create(subject, s);
		if (ind != null) {
		    indications.add(ind);
		}
	    }
	    RDFNode n = s.getObject();
	    if (n instanceof Resource) {
		if (!isValidInternalResource((Resource)n, overallModel, localURI)) {
		    // This is to get around problems with the name of imported URIs
		    if (overallModel.importsPredicates.contains(predicate)) {
			// This is an imported URI, so it probably won't match, so skip
		    } else if ((overallModel.onPropertyPredicates.contains(predicate)) ||
			       (overallModel.subPropertyOfPredicates.contains(predicate)) ||
			       (overallModel.samePropertyAsPredicates.contains(predicate)) )  {
			// This is an undefined property?
			if (DEBUG) System.out.println("Making undefined property explicit in model: "+(Resource)n);
			overallModel.defineGenericProperty((Resource)n);
			indications.add(new UndefinedPropertyIndication((Resource)n, s));
		    } else {
			Indication ind = UndefinedResourceIndication.create((Resource)n, s);
			if (ind != null) {
			    indications.add(ind);
			}
		    }
		}
	    }
	}
    }	    


    /**
     * Returns true if the resource is an unnamed class as genereated by RDF-API.
     */
    private boolean isGenid(Resource r, String namespace)
    {
	//System.out.println("isGenid("+r+", "+namespace+")");
	String s = r.toString();
	int  hashPosn = s.lastIndexOf('#');
	if ((hashPosn >= 0) && (s.regionMatches(hashPosn+1, "genid", 0, 5))) {
	    String numStr = s.substring(hashPosn+6);
	    try {
		int n = Integer.parseInt(numStr);
		//System.out.println("true");
		return(true);
	    } catch (NumberFormatException e) {
	    }
	}
	//System.out.println("false");
	return(false);
    }
	    
    /**
     * Returns false if the resource is in the namespace localURI but has no class type.
     * Resources in other namespaces or without a namespace produce a true result.
     */
    private boolean isValidInternalResource(Resource r, DAMLModelImpl m, String localURI) 
	throws RDFException
    {
	// Only checking resources in the local namespace
	if (r.getNameSpace() == null) {
	    return(true);	
	} else if (r.getNameSpace().startsWith(localURI)) {
	    // if it is a local resource, check to see if it has a type.
	    DAMLModelImpl.Node node = overallModel.getNode(r);
	    if (node == null) {
		// This should not happen
		if (DEBUG) {System.out.println("DEBUG: No Node for internal Resource: "+r);}
		return(false);
	    } else {
		// Get node types...
		Set types = node.getTypes();
		//System.out.println("--isValidInternalResource("+r+")");
		//System.out.println("types = "+types);
		if (types.size() == 0) {
		    if (isGenid(r, localURI)) {
			// Local genid### Resources without types are generated by RDF-API from <comment/> 
			// This suppresses generating an error.
			return(true);
		    } else {
			// Node has no type, so it is probably an undefined resource.
			return(false);
		    }
		}
	    }
	}
	// Anything else is not a internal resource
	return(true);
    }
    

    /**
     * Validate all nodes referenced by model
     */
    private void validateNodes(Model validateModel, DAMLModelImpl overallModel) 
    throws RDFException
    {
	boolean DEBUG = false;
	System.out.println("Validating Model Nodes...");
	// Determine what nodes to validate
	// Use all resources that appear as the subject of a statement.
	HashSet nodes = new HashSet();
	for(StmtIterator en = validateModel.listStatements(); en.hasNext();) {
	    Statement s = (Statement)en.next();
	    Resource subject = s.getSubject();
	    nodes.add(overallModel.getNode(subject));
	}

	// Validate nodes based on their type(s)
	for(Iterator i = nodes.iterator(); i.hasNext() && (Indication.indicationCount() < maxIndications); ) {
	    DAMLModelImpl.Node node = (DAMLModelImpl.Node)i.next();

	    if (DEBUG) System.out.println();
	    if (DEBUG) System.out.println("Checking node "+node);
	    
	    // Get the types of this Node
	    Set types = node.getTypes();

	    if (DEBUG) System.out.println("has types "+types);

	    if (types.size() == 0) {
		// Node has no type - it is probably an undefined resource, or a statement
		if (DEBUG) System.out.println("DEBUG: Node has no type: "+node);
	    }

	    boolean isClass = false;
	    boolean isProperty = false;
	    boolean isObjectProperty = false;
	    boolean isDatatypeProperty = false;
	    boolean isInstance = false;
	    boolean isRestriction = false;
	    boolean isObjectRestriction = false;
	    boolean isDatatypeRestriction = false;
	    boolean isTypedLiteral = false;
	    DAMLModelImpl.Node literalType = null;

	    int typecount = 0;
	    for (Iterator j = types.iterator(); j.hasNext(); ) {
		DAMLModelImpl.Node type = (DAMLModelImpl.Node)j.next();
		if (overallModel.classClasses.contains(type.resource) && !isClass) {
		    // Increase typecount if not already a class
		    isClass = true;
		    typecount++;
		} else if (overallModel.propertyClasses.contains(type.resource) && !isProperty) {
		    isProperty = true;
		    // Increase typecount if not already a property
		    if (!isDatatypeProperty && !isObjectProperty) typecount++;
		} else if (overallModel.datatypePropertyClasses.contains(type.resource) && !isDatatypeProperty) {
		    isDatatypeProperty = true;
		    // Increate typecount if not already a property, or if also a object property
		    if (!isProperty || isObjectProperty) typecount++;
		} else if (overallModel.objectPropertyClasses.contains(type.resource) && !isObjectProperty) {
		    isObjectProperty = true;
		    // Increase typecount if not already a property, or if also a datatype property
		    if (!isProperty || isDatatypeProperty) typecount++;
		} else if (overallModel.restrictionClasses.contains(type.resource) && !isRestriction) {
		    isRestriction = true;
		    typecount++;
		} else if (overallModel.objectRestrictionClasses.contains(type.resource)) {
		    isObjectRestriction = true;
		    typecount++;
		} else if (overallModel.datatypeRestrictionClasses.contains(type.resource)) {
		    isDatatypeRestriction = true;
		    typecount++;
		} else if (datatypeValidator.hasNamespace(type.resource.getNameSpace())) {
		    isTypedLiteral = true;
		    literalType = type;
		    typecount++;
		} else {
		    isInstance = true;
		}
	    }
	    
	    // It is ok to be a class and an instance.
	    //if ((typecount > 1) || (isInstance && (typecount > 0))) {
	    if (typecount > 1) {
		//System.out.println("INDICATION: Node of conflicting types.");
		indications.add(new ResourceIndication(node.resource, node.getTypeStatements(), "Resource "+node+" is an instance of multiple conflicting types.", "Type Conflict"));
	    } else if (isClass) {
		ValidateClassNode(node, overallModel);
	    } else if (isProperty || isDatatypeProperty || isObjectProperty) {
		ValidatePropertyNode(node, overallModel, isObjectProperty, isDatatypeProperty);
	    } else if (isInstance) {
		ValidateInstanceNode(node, overallModel, types);
	    } else if (isRestriction || isObjectRestriction || isDatatypeRestriction) {
		ValidateRestrictionNode(node, overallModel);
	    } else if (isTypedLiteral) {
		ValidateTypedLiteralNode(node, literalType, overallModel);
	    }
	}
    }

    private void ValidateClassNode(DAMLModelImpl.Node node, DAMLModelImpl overallModel) 
    throws RDFException
    {
	boolean DEBUG = false;
	if (DEBUG) System.out.println("---In ValidateClassNode");
	// Make sure type type is Class
	// Um, this is only called if the type is class...
	if (!node.isInstanceOfResources(overallModel.classClasses)) {
	    //System.out.println("INDICATION: class not of type class:"+node.getTypes());
	    indications.add(new ResourceIndication(node.resource, node.getTypeStatements(), "Resource "+node+" is not of type Class.", "Not a Class"));
	    return;
	}

	// check for specific class properties...
	Vector oneOf = new Vector();
	Vector complementOf = new Vector();
	Vector unionOf = new Vector();
	Vector intersectionOf = new Vector();
	Vector other = new Vector();
	Vector errStmts = new Vector();
	
	// For each statement
	for (StmtIterator e = node.outgoing.listStatements(); e.hasNext(); ) {
	    Statement s = (Statement)e.next();
	    Resource p = s.getPredicate();

	    // Gather the Class specific properties
	    if (overallModel.oneOfPredicates.contains(p)) {
		oneOf.add(s);
	    } else if (overallModel.complementOfPredicates.contains(p)) {
		complementOf.add(s);
	    } else if (overallModel.unionOfPredicates.contains(p)) {
		unionOf.add(s);
	    } else if (overallModel.intersectionOfPredicates.contains(p)) {
		intersectionOf.add(s);
	    } else {
		// Other statements
		other.add(s);
	    }
	}

	int count = 0;

	// Enumeration - Note: RDF-API does not parse List types
	if (oneOf.size() > 0) {
	    count++;
	    errStmts.addAll(oneOf);
	    // Values can be of any type of object.
	}

	// Complement
	if (complementOf.size() == 1) {
	    count++;
	    errStmts.addAll(complementOf);
	    // Must only be one, and it must be a class.
	    Statement s = (Statement)complementOf.firstElement();
	    Object o = (Object)s.getObject();
	    if (o instanceof Resource) {
		DAMLModelImpl.Node n = overallModel.getNode((Resource)o);
		if ((n != null) && (n.isClassNode())) {
		    // Complement of is a class - OK.
		} else {
		    // This is tested at the statement level (domain/range on complementOf)
		    //System.out.println("INDICATION: Complement of is not a Class:"+complementOf.firstElement());
		}
	    } else {
		// This is tested at the statement level (domain/range on complementOf)
		//System.out.println("INDICATION: Complement of is not a Class: "+complementOf.firstElement());
	    }
	} else if (complementOf.size() > 1) {
	    count++;
	    errStmts.addAll(complementOf);
	    //System.out.println("INDICATION: complement of more than one class");
	    indications.add(new ResourceIndication(node.resource, complementOf, "Class "+node+" is complement of more than one class.", "ComplementOf Multiple Classes"));
	}

	// Union - Note: RDF-API does not parse List types
	if (unionOf.size() > 0) {
	    count++;
	    errStmts.addAll(unionOf);
	}
	    
	// Intersection - Note: RDF-API does not parse List types
	if (intersectionOf.size() > 0) {
	    count++;
	    errStmts.addAll(intersectionOf);
	}

	if (count > 1) {
	    System.out.println("INDICATION: Class is of more than one special type (boolean/enum).");
	    indications.add(new ResourceIndication(node.resource, errStmts, "Class "+node+" is more than one special class type.", "Multiple Special Classes"));
	}

    }

    private void ValidatePropertyNode(DAMLModelImpl.Node propertyNode, DAMLModelImpl overallModel, boolean isObjectProperty, boolean isDatatypeProperty) 
	throws RDFException
    {
	boolean DEBUG = false;
	if (DEBUG) System.out.println("\n---In ValidatePropertyNode("+propertyNode+")");
	// For Object/Datatype Properties check
	// subPropertyOf, range, samePropertyAs, equivalentTo
	// Can we do something with inverseOf?

	// General Property Checking
	// domain must be a class
	// Get all domain statements for the propertyNode
	Collection domainStmts = propertyNode.getDomainStatements();
	// For each domain statement (getDomain only returns valid classes)...
	for (Iterator i = domainStmts.iterator(); i.hasNext() ; ) {
	    // ...verify that it is a class and not a datatype
	    Statement statement = (Statement) i.next();
	    Object object = statement.getObject();
	    if (DEBUG) System.out.println("domain = "+object);
	    if (object instanceof Resource) {
		DAMLModelImpl.Node node = overallModel.getNode((Resource)object, false);
		if (node.isClassNode()) {
		    // Domain is a Class - Valid
		} else {
// 		indications.add(new IncorrectDomainIndication(statement, 
// 							      org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.Class, 
// 							      node.getTypes()));
		    try {
			indications.add(new IncorrectDomainIndication(statement, 
								      (Resource)overallModel.DAMLVocabulary.getField("Class").get(null),
								      node.getTypes()));
		    } catch (NoSuchFieldException e2) {
			System.err.println("Error in ValidatePropertyNode:\n"+e2);
		    } catch (IllegalAccessException e3) {
			System.err.println("Error in ValidatePropertyNode:\n"+e3);
		    }
		}
	    } else {
		// Domain is not a Class - Invalid
// 		indications.add(new IncorrectDomainIndication(statement, 
// 							      org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.Class, 
// 							      null));
		try {
		    indications.add(new IncorrectDomainIndication(statement, 
								  (Resource)overallModel.DAMLVocabulary.getField("Class").get(null),
								  null));
		    } catch (NoSuchFieldException e2) {
			System.err.println("Error in ValidatePropertyNode:\n"+e2);
		    } catch (IllegalAccessException e3) {
			System.err.println("Error in ValidatePropertyNode:\n"+e3);
		    }
	    }
	}

	if (isObjectProperty) {
	    // Object Property Checking
	    if (DEBUG) System.out.println(" isObjectProperty");

	    // subPropertyOf must be a Property or ObjectProperty
	    Collection subPropStmts = propertyNode.getSubPropertyOfStatements();
	    for (Iterator i = subPropStmts.iterator(); i.hasNext(); ) {
		Statement statement = (Statement)i.next();
		Object object = statement.getObject();
		if (DEBUG) System.out.println("subPropertyOf = "+object);
		if (object instanceof Resource) {
		    DAMLModelImpl.Node node = overallModel.getNode((Resource)object, false);
		    if (node.isType(overallModel.datatypePropertyClasses)) {
			Vector tmp = new Vector();
			indications.add(new StatementIndication(statement, "An ObjectProperty can not be a subPropertyOf a DatatypeProperty."));
			//tmp.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.ObjectProperty);
			//tmp.add(org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL.Property);
			//indications.add(new IncorrectRangeIndication(statement, tmp, node.getTypes()));
		    }
		}		    
	    }

	    // range restrictions must be an object, not a datatype
	    // Statement validation issues an indication if the range is not a class.
	    // Just need to check that it is not a Datatype
	    Collection rangeStmts = propertyNode.getRangeStatements();
	    for (Iterator i = rangeStmts.iterator(); i.hasNext(); ) {
		Statement statement = (Statement)i.next();
		Object object = statement.getObject();
		if (DEBUG) System.out.println("range = "+object);
		if (object instanceof Resource) {
		    DAMLModelImpl.Node node = overallModel.getNode((Resource)object);
		    if (node.isDatatypeNode()) {
			indications.add(new StatementIndication(statement, "The range restriction for an ObjectProperty can not be a datatype."));
		    }
		}
	    }	    

	    // samePropertyAs must be an ObjectProperty
	    Collection samePropStmts = propertyNode.getSamePropertyAsStatements();
	    for (Iterator i = samePropStmts.iterator(); i.hasNext(); ) {
		Statement statement = (Statement)i.next();
		Object object = statement.getObject();
		if (DEBUG) System.out.println("samePropertyAs = "+object);
		if (object instanceof Resource) {
		    DAMLModelImpl.Node node = overallModel.getNode((Resource)object, false);
		    if (!node.isType(overallModel.objectPropertyClasses)) {
			indications.add(new StatementIndication(statement, "An ObjectProperty should be the samePropertyAs another ObjectProperty"));
		    }
		} else {
		    // object isn't a Property at all
		    indications.add(new StatementIndication(statement, "An ObjectProperty should be the samePropertyAs another ObjectProperty"));
		}
	    }

	    // equivalentTo must be an ObjectProperty
	    Collection equivPropStmts = propertyNode.getEquivalentToStatements();
	    for (Iterator i = equivPropStmts.iterator(); i.hasNext(); ) {
		Statement statement = (Statement)i.next();
		Object object = statement.getObject();
		if (DEBUG) System.out.println("equivalentTo = "+object);
		if (object instanceof Resource) {
		    DAMLModelImpl.Node node = overallModel.getNode((Resource)object, false);
		    if (!node.isType(overallModel.objectPropertyClasses)) {
			indications.add(new StatementIndication(statement, "An ObjectProperty should be equivalentTo another ObjectProperty"));
		    }
		} else {
		    // object isn't a property at all -- range indication instead?
		    indications.add(new StatementIndication(statement, "An ObjectProperty should be equivalentTo another ObjectProperty"));
		}
	    }


	} else if (isDatatypeProperty) {
	    // Datatype Property Checking
	    if (DEBUG) System.out.println(" isDatatypeProperty");

	    // subPropertyOf must be a Property or DatatypeProperty
	    Collection subPropStmts = propertyNode.getSubPropertyOfStatements();
	    for (Iterator i = subPropStmts.iterator(); i.hasNext(); ) {
		Statement statement = (Statement)i.next();
		Object object = statement.getObject();
		if (DEBUG) System.out.println("subPropertyOf = "+object);
		if (object instanceof Resource) {
		    DAMLModelImpl.Node node = overallModel.getNode((Resource)object, false);
		    if (node.isType(overallModel.objectPropertyClasses)) {
			indications.add(new StatementIndication(statement, "A DatatypeProperty can not be a subPropertyOf an ObjectProperty."));
		    }
		}		    
	    }

	    // range restrictions must be a datatype
	    // Statement validation catchs ranges that are not classes.
	    Collection rangeStmts = propertyNode.getRangeStatements();
	    for (Iterator i = rangeStmts.iterator(); i.hasNext(); ) {
		Statement statement = (Statement)i.next();
		Object object = statement.getObject();
		if (DEBUG) System.out.println("range = "+object);
		if (object instanceof Resource) {
		    DAMLModelImpl.Node node = overallModel.getNode((Resource)object);
		    if (node.isClassNode() && !node.isDatatypeNode()) {
			indications.add(new StatementIndication(statement, "The range restriction for a datatypeProperty must be a datatype."));
			System.out.println("TBD - Datatype Property "+propertyNode+" range is not a datatype: "+node);
			System.out.println("at statement "+statement);
		    }
		}
	    }	    

	    // samePropertyAs must be a DatatypeProperty
	    Collection samePropStmts = propertyNode.getSamePropertyAsStatements();
	    for (Iterator i = samePropStmts.iterator(); i.hasNext(); ) {
		Statement statement = (Statement)i.next();
		Object object = statement.getObject();
		if (DEBUG) System.out.println("samePropertyAs = "+object);
		if (object instanceof Resource) {
		    DAMLModelImpl.Node node = overallModel.getNode((Resource)object, false);
		    if (!node.isType(overallModel.datatypePropertyClasses)) {
			indications.add(new StatementIndication(statement, "An DatatypeProperty should be the samePropertyAs another DatatypeProperty"));
			System.out.println("TBD - Datatype Property "+propertyNode+" can only be the samePropertyAs a datatype property.");
			System.out.println("at statement "+statement);
		    }
		} else {
		    indications.add(new StatementIndication(statement, "A DatatypeProperty should be the samePropertyAs another DatatypeProperty"));
		}		    
	    }

	    // equivalentTo must be a DatatypeProperty
	    Collection equivPropStmts = propertyNode.getEquivalentToStatements();
	    for (Iterator i = equivPropStmts.iterator(); i.hasNext(); ) {
		Statement statement = (Statement)i.next();
		Object object = statement.getObject();
		if (DEBUG) System.out.println("equivalentTo = "+object);
		if (object instanceof Resource) {
		    DAMLModelImpl.Node node = overallModel.getNode((Resource)object, false);
		    if (!node.isType(overallModel.datatypePropertyClasses)) {
			indications.add(new StatementIndication(statement, "A DatatypeProperty should be equivalentTo another DatatypeProperty"));
			System.out.println("TBD - Datatype Property "+propertyNode+" can only be equivalentTo a datatype property.");
			System.out.println("at statement "+statement);
		    }
		} else {
		    indications.add(new StatementIndication(statement, "A DatatypeProperty should be equivalentTo another DatatypeProperty"));
		}
	    }



	} else {
	    // General Property Checking
	    // subPropertyOf must be a Property
	    Collection subPropStmts = propertyNode.getSubPropertyOfStatements();
	    for (Iterator i = subPropStmts.iterator(); i.hasNext(); ) {
		Statement statement = (Statement)i.next();
		Object object = statement.getObject();
		if (DEBUG) System.out.println("subPropertyOf = "+object);
		if (object instanceof Resource) {
		    DAMLModelImpl.Node node = overallModel.getNode((Resource)object, false);
		    if (node.isType(overallModel.datatypePropertyClasses) || 
			node.isType(overallModel.objectPropertyClasses)) {
			indications.add(new StatementIndication(statement, "Property can not be a subPropertyOf a DatatypeProperty or ObjectProperty."));
			System.out.println("TBD - Property "+propertyNode+" must be a subPropertyOf a property.");
			System.out.println("at statement "+statement);
		    }
		}		    
	    }

	}
	
    }

    private void ValidateRestrictionNode(DAMLModelImpl.Node node, DAMLModelImpl overallModel)
    {
	boolean DEBUG = false;
	if (DEBUG) System.out.println("---In ValidateRestrictionNode");
	// validate the general restriction properties
	// Must have an onProperty value
	// Datatype Restrictions are on Datatype Properties
	// Object Restrictions are on Object Properties
	// check toClass, hasValue, hasClass, hasClassQ, 
    }

    private void ValidateTypedLiteralNode(DAMLModelImpl.Node node, DAMLModelImpl.Node type, DAMLModelImpl overallModel)
	throws RDFException
    {
	boolean DEBUG = false;
	if (DEBUG) System.out.println("---In ValidateTypedLiteralNode");
	// make sure the literal value matches the datatype definition
	Set values = node.getValues();
	int size = values.size();
	if (size == 1) {
	    RDFNode val = (RDFNode)(values.iterator().next());
	    if (val instanceof Literal) {
		try {
		    String value = ((Literal)val).getString();
		    if (!datatypeValidator.isAValidValue(value, type.resource.getNameSpace(),
							 type.resource.getLocalName())) {
			// value does not fit datatype
			// System.out.println("TBD ERROR: value "+value+" does not match DataType "+type.resource.getURI());
			HashSet stmts = new HashSet(node.getTypeStatements());
			stmts.addAll(node.getValueStatements());
			indications.add(new TypedLiteralMismatchIndication(type.resource, (Literal)val, stmts));
		    }
		} catch (UnknownSimpleTypeException uste) {
		    // Datatype is not defined - reported as undefined resource indication elsewhere
		    // System.out.println(uste);
		}
		    
	    } else {
		// Error: typed literal should be a literal
		// System.out.println("TBD ERROR: Typed Literal value ("+val+") should be a Literal");
		HashSet stmts = new HashSet(node.getTypeStatements());
		stmts.addAll(node.getValueStatements());
		indications.add(new LiteralExpectedIndication(type.resource, val, stmts));
	    }	
	} else if (size == 0) {
	    // Typed literal has no value.
	    //System.out.println("TBD ERROR: datatype instance "+node+" has no value.");
	    HashSet stmts = new HashSet(node.getTypeStatements());
	    indications.add(new TypedLiteralValueIndication(type.resource, values, stmts));
	} else {
	    // Typed literal has more than one value.
	    //System.out.println("TBD ERROR: datatype instance "+node+" has more than one value: "+values);
	    HashSet stmts = new HashSet(node.getTypeStatements());
	    stmts.addAll(node.getValueStatements());
	    indications.add(new TypedLiteralValueIndication(type.resource, values, stmts));
	}
    }


    private void ValidateInstanceNode(DAMLModelImpl.Node node, DAMLModelImpl overallModel, Set types) 
    throws RDFException
    {
	boolean DEBUG = false;
	if (DEBUG) System.out.println("\n---In ValidateInstanceNode "+node);
	// For each type
	for (Iterator i = types.iterator(); i.hasNext(); ) {
	    DAMLModelImpl.Node type = (DAMLModelImpl.Node)i.next();
	    if (DEBUG) System.out.println("   Type: "+type);

	    // Get any property restrictions for this type class.
	    Set restrictions = type.getRestrictions(false);

	    // For each restriction
	    for (Iterator j = restrictions.iterator(); j.hasNext(); ) {
		DAMLModelImpl.Node r = (DAMLModelImpl.Node)j.next();
		if (DEBUG) System.out.println("  -restriction Node: "+r);

		// Create a restriction object
		DAMLModelImpl.Restriction restriction = overallModel.getRestriction(r);
		if (DEBUG) System.out.println("   restriction obj: "+restriction);
		if (DEBUG) System.out.println("   onProperty: "+restriction.onProperty);
	       
		if (restriction.onProperty != null) {
		    // Get all statements for this property from this instance Node
		    Vector properties = DAMLModelImpl.getAllStatements(node.outgoing, restriction.onProperty.resource);
		    if (DEBUG) System.out.println("   node properties: "+properties);

		    // Check toClass restriction (if exists)
		    if (restriction.toClass != null) {
			// All the statement ranges must be of this class
			for (Enumeration e = properties.elements(); e.hasMoreElements(); ) {
			    Statement s = (Statement)e.nextElement();
			    if (!overallModel.isInstanceOf(s.getObject(), restriction.toClass)) {
				// InvalidRangeIndication
				Set t = null;
				if (s.getObject() instanceof Resource) {
				    t = overallModel.getNode((Resource)s.getObject()).getTypes();
				} 
				indications.add(new IncorrectRangeIndication(s, restriction.toClass, t));
			    }
			}
		    }

		    // Check hasValue restriction (if exists)
		    if (restriction.hasValue != null) {
			if (DEBUG) System.out.println("  -hasValue "+restriction.hasValue.getClass()+" -- "+restriction.hasValue);
			// One of the values must be this
			boolean exists = false;
			for (Enumeration e = properties.elements(); e.hasMoreElements(); ) {
			    Statement s = (Statement)e.nextElement();
			    if (DEBUG) System.out.println("    -"+s.getObject().getClass()+" -- "+s.getObject());
			    if ((restriction.hasValue instanceof DAMLModelImpl.Node) &&
				((DAMLModelImpl.Node)restriction.hasValue).equals(s.getObject())) {
				if (DEBUG) System.out.println("MATCHED");
				exists = true;
			    } else if ((restriction.hasValue instanceof RDFNode) &&
				       ((RDFNode)restriction.hasValue).equals(s.getObject())) {
				if (DEBUG) System.out.println("MATCHED");
				exists = true;
			    }
			    
			}
			if (!exists) {
			    // hasValueIndication
			    indications.add(new HasValueIndication(node, type, restriction, properties));
			}
		    }

		    // Check hasClass restriction (if exists)
		    if (restriction.hasClass != null) {
			// At least one value must be of the class (or datatype)
			boolean exists = false;
			for (Enumeration e = properties.elements(); e.hasMoreElements(); ) {
			    Statement s = (Statement)e.nextElement();
			    if (overallModel.isInstanceOf(s.getObject(), restriction.hasClass)) {
				exists = true;
			    }
			}			
			if (!exists) {
			    // hasClassIndication
			    indications.add(new HasClassIndication(node, type, restriction, properties));
			}
		    }

		    // Check cardinality restrictions (if exist)
		    // Note: This may not correctly test for N distinct values
		    if ( ((restriction.minCardinality >= 0) &&
			  (restriction.minCardinality > properties.size())) ||
			 ((restriction.maxCardinality >= 0) &&
			  (restriction.maxCardinality < properties.size())) ) {
			// CardinalityIndication
			indications.add(new CardinalityIndication(node, type, restriction, properties));
		    }

		    // Check cardinalityQ restrictions (if exist)
		    // Note: This may not correctly test for N distinct values
		    if ((restriction.hasClassQ != null) &&
			((restriction.minCardinalityQ >= 0) ||
			 (restriction.maxCardinalityQ >= 0))) {

			// Determine the distinct values from the named class
			if (DEBUG) System.out.println("  -cardinalityQ Restriction");
			HashSet values = new HashSet();
			HashSet stmts = new HashSet();
			for (Enumeration e = properties.elements(); e.hasMoreElements(); ) {
			    Statement s = (Statement)e.nextElement();
			    if (DEBUG) System.out.println("    "+s.getObject());
			    if (overallModel.isInstanceOf(s.getObject(), restriction.hasClassQ)) {
				values.add(s.getObject());
				stmts.add(s);
			    }
			}
			if (DEBUG) System.out.println("   statements: "+values);

			if ( ((restriction.minCardinalityQ >= 0 ) &&
			      (restriction.minCardinalityQ > values.size())) ||
			     ((restriction.maxCardinalityQ >= 0) &&
			      (restriction.maxCardinalityQ < values.size())) ) {
			    // CardinalityQIndication
			    indications.add(new CardinalityQIndication(node, type, restriction, stmts));
			    if (DEBUG) System.out.println("   MATCHED");
			}
		    }
			
		}
	    }
	}
    }


    /**
     * Adds RDF Model to the set of models used to validate against.
     */
    public void addModel(Model model) 
	throws RDFException
    {
	overallModel.addModel(model);
    }

    /**
     * Load URL to create a RDF Model and add that to the set of models to validate against.
     */
    public void addURL(java.net.URL url, boolean recurse)
	throws RDFException
    {
	addURL(url.toString(), recurse);
    }

    public void addURL(String url, boolean recurse)
	throws RDFException
    {
	Model m = new ModelMem();
// 	Model m = factory.createModel();
// 	parser = (SiRPAC)factory.createParser();
	ParseHelper p = new ParseHelper(preferences, overallModel.importsPredicates, datatypeValidator);
	p.parseURI(url, m, recurse);

	addModel(m);
    }


    /**
     * Starts the validation process.
     */
    public void validate(Model validateModel, String uri)
	throws RDFException
    {
	// Keep track of the uri that we are validating
	currentURI = uri;

	// Determine if there are any statements in the validate model
	if (validateModel.size() == 0) {
	    Indication i = new NotRDFIndication(currentURI);
	    indications.add(i);
	} else {

	    // Validate the existance of referenced nodes that are outside
	    // if the namespace being validated.
	    validateReferencedResources(validateModel, overallModel);

	    if (Indication.indicationCount() < maxIndications) {

		// Place the validate model into the overall model
		overallModel.addModel(validateModel);

		// Validate resources that are in the namespace of the validateModel
		validateInternalResources(validateModel, overallModel);

		// Validate internal references
		// Check to see if every local resource has a type
		// This is done as part of Node Validation

		if (Indication.indicationCount() < maxIndications) {
		    // Validate Statements
		    validateStatements(validateModel, overallModel);
		}

		if (Indication.indicationCount() < maxIndications) {
		    // Validate Nodes
		    validateNodes(validateModel, overallModel);
		}

	    }
	}
    }
}
