// $Id: ValidatorModelMem.java,v 1.1 2003/03/12 21:31:10 drager Exp $

package org.daml.validator;

import com.hp.hpl.mesa.rdf.jena.mem.ModelMem;
import com.hp.hpl.mesa.rdf.jena.model.RDFException;
import org.daml.validator.indications.Indication;
import java.util.Vector;

    /**
     * Jena model with a reader that records errors/warnings as Validator indications
     * and suppresses output of ARP warnings.
     */

    public class ValidatorModelMem
	extends ModelMem
    {
	private final boolean report;
	private final Vector indications = new Vector();

	ValidatorModelMem(boolean reportIndications)
	{
	    super();
	    report = reportIndications;

	    readerFactory = new com.hp.hpl.mesa.rdf.jena.common.RDFReaderFImpl()
		{
		    public com.hp.hpl.mesa.rdf.jena.model.RDFReader getReader()
			throws RDFException
		    {
			com.hp.hpl.mesa.rdf.jena.model.RDFReader retval = super.getReader();
			com.hp.hpl.jena.rdf.arp.JenaReader reader = (com.hp.hpl.jena.rdf.arp.JenaReader) retval;
			reader.setErrorHandler(new com.hp.hpl.mesa.rdf.jena.model.RDFErrorHandler()
			    {
				void indicate(Exception e,
					      int level)
				{
				    String message = e.toString();
				    if (e instanceof org.xml.sax.SAXParseException)
					{
					    // suppress the warnings about RDF datatyping not supported in Jena1
					    if (e instanceof com.hp.hpl.jena.rdf.arp.ParseException) {
						if (((com.hp.hpl.jena.rdf.arp.ParseException)e).getErrorNumber() 
						    == com.hp.hpl.jena.rdf.arp.ARPErrorNumbers.WARN_NOT_SUPPORTED &&
						    (e.toString().indexOf("RDF datatyping not supported") >= 0) 
						    )
						    return;
					    }
					    org.xml.sax.SAXParseException sax = (org.xml.sax.SAXParseException) e;
					    message += (" at line " + sax.getLineNumber() + ", column " + sax.getColumnNumber());
					}
				    if (report) {
					Indication indication = new Indication(message); // XXX - use more specific Indication subclass including separate line and column
					indication.setLevel(level);
					indications.add(indication);
				    }
				}

				public void error(Exception e)
				{
				    indicate(e, Indication.ERROR);
				}
				
				public void fatalError(Exception e)
				{
				    indicate(e, Indication.ERROR);
				}
				
				public void warning (Exception e)
				{
				    indicate(e, Indication.WARNING);
				}
			    });

			return retval;
		    }
		};
	}

	public Vector getIndications() 
	{
	    return indications;
	}

	
    }

