// $Id: TypedLiteralMismatchIndication.java,v 1.2 2002/08/05 19:03:09 drager Exp $

package org.daml.validator.indications;

import com.hp.hpl.mesa.rdf.jena.model.*;
import java.util.*;

/**
 * Indication used for typed literals whose value does not match the datatype.
 */
public class TypedLiteralMismatchIndication extends Indication {

    /**
     * The datatype of the typed literal.
     */
    Resource datatype;

    /**
     * The value of the typed literal.
     */
    Literal value;

    /**
     * The type and value statements used for the location
     */
    Collection statements;

    public static int defaultLevel = Indication.ERROR;


    /**
     * Constructor function.
     * @param datatype The datatype 
     * @param value    The value
     * @param statements The datatype type and value statements (for location)
     */
    public TypedLiteralMismatchIndication(Resource datatype, Literal value, Set statements)
    { 
	super();
	this.datatype = datatype;
	this.value = value;
	this.statements = statements;
	type = "Typed Literal Mismatch";
	level = defaultLevel;
    }


    public String getMessage()
    {
	String m = "";
	m = "Datatype value "+value+" is not of type "+datatype+".";
	return m;
    }
    

    public String getLocation()
    {
	return(getLocation(statements));
    }

    public String getLocationXML()
	throws RDFException
    {
	return getLocationXML(statements);
    }

}
