// $Id: URIIndication.java,v 1.3 2002/08/05 19:03:09 drager Exp $

package org.daml.validator.indications;

import java.util.Stack;
import com.hp.hpl.mesa.rdf.jena.model.*;

/**
 * Generic Indication for URI based Indications.
 */
public class URIIndication extends Indication {
   
    /**
     * URI the indication applies to.
     */
    public String uri;

    /**
     * The call stack as an array, from the indication URI to the file being verified.
     */
    public String stack[];

    public static int defaultLevel = Indication.WARNING;

    public URIIndication(String uri, Stack callStack) 
    {
	super();
	level = defaultLevel;
	type = "URI Indication";
	this.uri = uri;
	if (callStack == null) {
	    this.stack = new String[0];
	    //this.stack[0] = uri;
	} else {
	    this.stack = stackToArray(callStack);
	}
    }

    private String[] stackToArray(Stack stack) 
    {
	String[] result = new String[stack.size()];
	for (int i = 0; i < stack.size(); ) {
	    result[i] = (String)stack.elementAt(i);
	    i++;
	}
	return(result);
    }
    
    public String getMessage()
    {
	String m = "";
	m += "Generic URI Indication for "+uri;
	return(m);
    }

    public String getLocation()
    {
	if ((stack == null) || (stack.length == 0)) {
	    return(uri);
	}

	String indent = "";
	String location = "";
	for (int i = 0; i < stack.length; ) {
	    location += indent+stack[i++]+"\n";
	    indent += "  ";
	}
	location += indent+uri+"\n";
	return(location);
    }

    public String getLocationXML()
	throws RDFException
    {
	if ((stack == null) || (stack.length == 0)) {
	    return(getLocationXML(uri));
	} else {
	    return(getLocationXMLStack(stack, uri));
	}
    }

    public String getLocationXMLStack (String[] stack, String uri) 
    {
	String xml = "<location>\n";
	xml += "<stack>\n";
	for (int i = 0; i < stack.length; ) {
	    xml += "<uri>"+ stack[i++] +"</uri>\n";
	}
	xml += "<uri>"+uri+"</uri>\n";
	xml += "</stack>\n";
	xml += "</location>\n";
	return(xml);
    }
	    
}
